/*
 * First attempt at a shared counter.
 * Threads alice and bob each increment the same counter
 * many times.
 *
 * This version of the program has a data race on counter.
 * Alice and Bob are reading and writing counter without
 * synchronizing with each other.
 *
 * This bug may result in lost updates to the counter!
 *
 * CSE 374, 2015-03 BM
 */

#include <thread>
#include <mutex>
#include <iostream>

using namespace std;

int counter = 0;

const int N = 1024*1024;

void alice_f() {
  for (int i=0; i<N; i++) {
    // there are two major problems with this line.
    // 1. counter++ is not atomic. It may help to think of it really as:
    //
    //    int temp = counter
    //    temp = temp + 1
    //    counter = temp
    //   
    //    Alice and Bob may interleave these operations.
    //    So, for example, Alice and Bob may both read 100, add 1, and both write 101.
    // 2. More advanced: there is no synchronization between Alice and Bob, so
    //    the compiler is actually allowed to break the program 
    //    (for example, make a private version of counter for both alice and bob rather than a single global one)
    counter++;
  }
}

void bob_f() {
  for (int i=0; i<N; i++) {
    // see above why this line is buggy
    counter++;
  }
}

int main() {
  thread alice(alice_f);   
  thread bob(bob_f); 

  alice.join();
  bob.join();
  // these joins are synchronization between alice/bob and the main thread,
  // so now the main thread can see the result of the counter increments.

  cout << "expected: " << N*2 << endl;
  cout << "actual: " << counter << endl;
}
